"""
Mapping of col to tyoes


"""
DOCUMENT = "document"
TOKEN = "token"
WORDPIECE = "wordpiece"
WORD_EMBEDDINGS = "word_embeddings"
SENTENCE_EMBEDDINGS = "sentence_embeddings"
CATEGORY = "category" # Multi Class etc
DATE = "date"
ENTITY = "entity"
SENTIMENT = "sentiment"
POS = "pos"
CHUNK = "chunk"
NAMED_ENTITY = "named_entity"
NEGEX = "negex"
DEPENDENCY = "dependency"
LABELED_DEPENDENCY = "labeled_dependency"
LANGUAGE = "language"
KEYWORD = "keyword"
DUMMY = "dummy"