"""
Mapping of col to tyoes


"""
DOCUMENT = "document"
TOKEN = "token"
WORDPIECE = "wordpiece"
WORD_EMBEDDINGS = "word_embeddings"
SENTENCE_EMBEDDINGS = "sentence_embeddings"
CATEGORY = "category" # Multi Class etc
DATE = "date"
SENTIMENT = "sentiment"
POS = "pos"
CHUNK = "chunk" # Multiple Token text piece. i.e. extracted entity
NAMED_ENTITY = "named_entity" # exxtracted IOB
ENTITY = "entity" # converted entity, pretty
NEGEX = "negex" # ?
UNDEPENDENCY = "unlabeled_dependency"
LABELED_DEPENDENCY = "labeled_dependency"
LANGUAGE = "language"
KEYWORD = "keyword"
